import hashlib
import os
import shutil
import uuid
from pathlib import Path
from typing import Any
from typing import Literal
from typing import Union

from pydantic import model_validator

from laktory._logger import get_logger
from laktory.models.basemodel import BaseModel
from laktory.models.pipeline.pipelinechild import PipelineChild
from laktory.polars import PolarsLazyFrame
from laktory.polars import is_polars_dataframe
from laktory.spark import SparkDataFrame
from laktory.spark import is_spark_dataframe
from laktory.typing import AnyDataFrame

logger = get_logger(__name__)


class DataSinkMergeCDCOptions(BaseModel):
    """
    Options for merging a change data capture (CDC).

    They are also used to build the target using `apply_changes` method when
    using Databricks DLT.

    Attributes
    ----------
    delete_where:
        Specifies when a CDC event should be treated as a DELETE rather than
        an upsert.
    end_at_column_name:
        When using SCD type 2, name of the column storing the end time (or
        sequencing index) during which a row is active. This attribute is not
        used when using Databricks DLT which does not allow column rename.
    exclude_columns:
        A subset of columns to exclude in the target table.
    ignore_null_updates:
        Allow ingesting updates containing a subset of the target columns.
        When a CDC event matches an existing row and ignore_null_updates is
        `True`, columns with a null will retain their existing values in the
        target. This also applies to nested columns with a value of null. When
        ignore_null_updates is `False`, existing values will be overwritten
        with null values.
    include_columns:
        A subset of columns to include in the target table. Use
        `include_columns` to specify the complete list of columns to include.
    order_by:
        The column name specifying the logical order of CDC events in the
        source data. Used to handle change events that arrive out of order.
    primary_keys:
        The column or combination of columns that uniquely identify a row in
        the source data. This is used to identify which CDC events apply to
        specific records in the target table.
    scd_type:
        Whether to store records as SCD type 1 or SCD type 2.
    start_at_column_name:
        When using SCD type 2, name of the column storing the start time (or
        sequencing index) during which a row is active. This attribute is not
        used when using Databricks DLT which does not allow column rename.

    References
    ----------
    - [How to Implement SCD 2 using Delta Table](https://iterationinsights.com/article/how-to-implement-slowly-changing-dimensions-scd-type-2-using-delta-table/)
    - [Change Data Capture with Databricks DLT](https://docs.databricks.com/en/delta-live-tables/python-ref.html#change-data-capture-with-python-in-delta-live-tables)
    """

    # apply_as_truncates: Union[str, None] = None
    delete_where: str = None
    end_at_column_name: str = "__end_at"
    exclude_columns: list[str] = None
    ignore_null_updates: bool = False
    include_columns: list[str] = None
    order_by: str = None
    primary_keys: list[str] = None
    scd_type: Literal[1, 2] = 1
    start_at_column_name: str = "__start_at"
    # track_history_columns: Union[list[str], None] = None
    # track_history_except_columns: Union[list[str], None] = None
    _parent: Any = None
    _source_schema: Any = None
    _source_columns: list[str] = None

    @model_validator(mode="after")
    def validate_scd_type(self) -> Any:
        if self.scd_type == 2:
            if self.order_by is None:
                raise ValueError(
                    "SCD Type 2 merge requires specific of `order_by` attribute."
                )

        return self

    @model_validator(mode="after")
    def selected_columns(self) -> Any:
        if self.include_columns and self.exclude_columns:
            raise ValueError(
                "`include_columns` and `exclude_columns` attributes are mutually exclusive."
            )

        if self.exclude_columns and self.order_by:
            if self.order_by in self.exclude_columns:
                raise ValueError(
                    f"`order_by` '{self.order_by}' can't be excluded as it will prevent proper merge of out-of-sequence records."
                )

        return self

    # ----------------------------------------------------------------------- #
    # Sink                                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def sink(self):
        return self._parent

    @property
    def target_name(self):
        if self._parent and "TableDataSink" in str(type(self._parent)):
            return self._parent.full_name
        return None

    @property
    def target_path(self):
        if self._parent and "FileDataSink" in str(type(self._parent)):
            return self._parent.path
        return None

    @property
    def target_id(self):
        return self.target_name or self.target_path

    # ----------------------------------------------------------------------- #
    # CDC Columns                                                             #
    # ----------------------------------------------------------------------- #

    @property
    def index(self):
        return self.order_by

    @property
    def start_at(self):
        return self.start_at_column_name

    @property
    def end_at(self):
        return self.end_at_column_name

    @property
    def index_fist(self):
        return self.index + "_first"

    @property
    def hash_keys(self):
        return "__hash_keys"

    @property
    def hash_cols(self):
        return "__hash_cols"

    @property
    def source_columns(self):
        return [f.name for f in self._source_schema]

    @property
    def update_columns(self):
        if self.include_columns:
            columns = [c for c in self.include_columns]
        else:
            columns = [c for c in self.source_columns if c not in self.primary_keys]
            if self.exclude_columns:
                columns = [c for c in columns if c not in self.exclude_columns]
        return columns

    @property
    def extra_columns(self):
        cols = [self.hash_keys]
        if self.scd_type == 2:
            cols += [self.start_at, self.end_at, self.hash_cols]
        return cols

    @property
    def write_columns(self):
        return self.primary_keys + self.update_columns + self.extra_columns

    @property
    def index_type(self):
        return [
            field.dataType
            for field in self._source_schema.fields
            if field.name == self.index
        ][0]

    @property
    def source_delete_where(self):
        return self._add_alias(self.delete_where)

    # ----------------------------------------------------------------------- #
    # Methods                                                                 #
    # ----------------------------------------------------------------------- #

    @staticmethod
    def _add_alias(expr, prefix="source"):
        operators = ["=", ">", "<", "!", "*", "+", "-", "/", ","]

        new_expr = expr
        for o in operators:
            new_expr = new_expr.replace(o, " ")

        candidates = [c for c in new_expr.split(" ") if c != ""]

        # Define other exclusions for SQL keywords or constants
        exclusions = [
            "TRUE",
            "FALSE",
            "CASE",
            "WHEN",
            "THEN",
            "SELECT",
            "FROM",
            "WHERE",
            "AND",
            "OR",
            "AS",
            "END",
            "ELSE",
            "NOT",
            "NULL",
        ]

        # Filter out identifiers that are constants or keywords
        column_names = []
        for c in candidates:
            if c.upper() in exclusions:
                continue
            if c.startswith("'"):
                continue
            try:
                float(c)
                continue
            except ValueError:
                pass

            column_names += [c]

        new_expr = expr
        for c in column_names:
            if f"{prefix}." not in c:
                new_expr = new_expr.replace(c, f"{prefix}.{c}")

        return new_expr

    def _init_target(self, source):
        import pyspark.sql.types as T

        spark = source.sparkSession
        logger.info(f"Merge target not found. Creating empty table at {self.target_id}")
        schema = source.select(self.primary_keys + self.update_columns).schema
        schema.add(T.StructField(self.hash_keys, T.StringType(), True))
        if self.scd_type == 2:
            schema.add(T.StructField(self.hash_cols, T.StringType(), True))
            schema.add(T.StructField(self.start_at, self.index_type, True))
            schema.add(T.StructField(self.end_at, self.index_type, True))

        df = spark.createDataFrame(data=[], schema=schema)

        writer = df.write.format("delta").mode("OVERWRITE")
        if self.target_path:
            writer.save(self.target_path)
        else:
            writer.saveAsTable(self.target_name)

    def _execute(self, source: SparkDataFrame):
        import pyspark.sql.functions as F
        from delta.tables import DeltaTable
        from pyspark.sql import Window

        spark = source.sparkSession

        logger.info(
            f"Executing merge on {self.target_id} with primary keys {self.primary_keys} and scd type {self.scd_type}"
        )

        if self.delete_where:
            logger.info(f"with delete on {self.delete_where}")

        # Add internal columns
        source = source.withColumn(
            self.hash_keys, F.lit(F.sha2(F.concat_ws("~", *self.primary_keys), 256))
        )
        if self.scd_type == 2:
            source = (
                source.withColumn(self.start_at, F.col(self.index))
                .withColumn(self.end_at, F.lit(None).cast(self.index_type))
                .withColumn(
                    self.hash_cols,
                    F.lit(F.sha2(F.concat_ws("~", *self.update_columns), 256)),
                )
            )

        # Process History
        if self.index:
            w = Window.partitionBy(*self.primary_keys).orderBy(F.desc(self.index))
            source = source.withColumn("_row_number", F.row_number().over(w))
            if self.scd_type == 1:
                # Drop Duplicates
                logger.info(
                    f"Dropping duplicates using {self.primary_keys} and '{self.order_by}' as sequencing index"
                )
                source = source.filter(F.col("_row_number") == 1)
            elif self.scd_type == 2:
                # Assign previous index to ends_at
                w2 = Window.partitionBy(*self.primary_keys)
                source = source.withColumn(self.end_at, F.lag(self.index, 1).over(w))
                source = source.withColumn(self.index_fist, F.min(self.index).over(w2))
            source = source.drop("_row_number")
        else:
            logger.info(f"Dropping duplicates using {self.primary_keys}")
            source = source.drop_duplicates(subset=self.primary_keys)

        # Read target
        if self.target_path:
            table_target = DeltaTable.forPath(spark, self.target_path)
        else:
            table_target = DeltaTable.forName(spark, self.target_name)

        if self.scd_type == 1:
            if self.delete_where:
                delete_condition = F.coalesce(
                    F.expr(self.source_delete_where), F.lit(False)
                )
                not_delete_condition = ~delete_condition

            # Define merge
            merge = table_target.alias("target").merge(
                source.alias("source"),
                condition=f"source.{self.hash_keys} = target.{self.hash_keys}",
            )

            # Update
            _set = {f"target.{c}": f"source.{c}" for c in self.update_columns}
            if self.ignore_null_updates:
                _set = {
                    f"target.{c}": F.coalesce(
                        F.col(f"source.{c}"), F.col(f"target.{c}")
                    ).alias(c)
                    for c in self.update_columns
                }

            condition = None
            if self.delete_where:
                condition = not_delete_condition
            if self.order_by:
                _condition = F.expr(f"source.{self.order_by} > target.{self.order_by}")
                if condition is None:
                    condition = _condition
                else:
                    condition = condition & _condition

            merge = merge.whenMatchedUpdate(set=_set, condition=condition)

            # Insert
            condition = None
            if self.delete_where:
                condition = not_delete_condition
            merge = merge.whenNotMatchedInsert(
                values={f"target.{c}": f"source.{c}" for c in self.write_columns},
                condition=condition,
            )

            # Delete
            if self.delete_where:
                merge = merge.whenMatchedDelete(condition=delete_condition)

            logger.info("Executing merge...")
            merge.execute()

        elif self.scd_type == 2:
            delete_condition = F.coalesce(F.expr(self.delete_where), F.lit(False))
            not_delete_condition = ~delete_condition

            # Only select rows that have been updated
            if self.target_path:
                target = spark.read.format("delta").load(self.target_path)
            else:
                target = spark.read.table(self.target_name)
            upsert_or_delete = source.withColumn("__to_delete", delete_condition).join(
                other=target.withColumn("__to_delete", F.lit(False)),
                on=[self.hash_cols, self.hash_keys, "__to_delete"],
                how="leftanti",
            )

            # Merge
            condition = F.expr(f"source.{self.hash_keys} = target.{self.hash_keys}")
            condition = condition & F.expr(f"target.{self.end_at} IS NULL")
            merge = table_target.alias("target").merge(
                upsert_or_delete.filter(F.col(self.end_at).isNull()).alias("source"),
                condition=condition,
            )

            # Expire the current record
            _set = {f"target.{self.end_at}": f"source.{self.index_fist}"}
            merge = merge.whenMatchedUpdate(set=_set)
            #
            # # TODO: Review if required
            # if not self.delete_where:
            #     _set = {f"target.{self.end_at}": f"source.{self.order_by}"}
            #     merge = merge.whenMatchedUpdate(set=_set)
            # else:
            #     where = F.expr(self.source_delete_where)
            #     # deleting
            #     # _set = {f"target.{self.end_at}": "NULL"}
            #     _set = {f"target.{self.end_at}": f"source.{self.order_by}"}
            #     merge = merge.whenMatchedUpdate(set=_set, condition=where)
            #
            #     # updating
            #     _set = {f"target.{self.end_at}": f"source.{self.order_by}"}
            #     merge = merge.whenMatchedUpdate(set=_set, condition=~where)

            logger.info("Executing merge...")
            merge.execute()

            # Append rows
            upsert = upsert_or_delete
            if self.delete_where:
                upsert = upsert.filter(not_delete_condition)
            writer = (
                upsert.select(self.write_columns).write.mode("APPEND").format("DELTA")
            )
            logger.info("Appending new rows...")
            if self.target_path:
                writer.save(self.target_path)
            else:
                writer.saveAsTable(self.target_name)

        else:
            raise ValueError(f"SCD Type {self.scd_type} is not supported.")

    def execute(self, source: SparkDataFrame):
        """
        Merge source into target delta from sink

        Parameters
        ----------
        source:
            Source DataFrame to merge into target (sink).
        """

        from delta.tables import DeltaTable

        self._source_schema = source.schema
        spark = source.sparkSession

        if self.target_path:
            if not DeltaTable.isDeltaTable(spark, self.target_path):
                self._init_target(source)
        else:
            try:
                spark.catalog.getTable(self.target_name)
            except Exception:
                self._init_target(source)

        if source.isStreaming:
            if self.sink is None:
                raise ValueError("Sink value required to fetch checkpoint location.")

            if self.sink and self.sink._checkpoint_location is None:
                raise ValueError(
                    f"Checkpoint location not specified for sink '{self.sink}'"
                )

            query = (
                source.writeStream.foreachBatch(
                    lambda batch_df, batch_id: self._execute(source=batch_df)
                )
                .trigger(availableNow=True)
                .options(
                    checkpointLocation=self.sink._checkpoint_location,
                )
                .start()
            )
            query.awaitTermination()

        else:
            self._execute(source=source)


class BaseDataSink(BaseModel, PipelineChild):
    """
    Base class for building data sink

    Attributes
    ----------
    is_primary:
        A primary sink will be used to read data for downstream nodes when
        moving from stream to batch. Don't apply for quarantine sinks.
    is_quarantine:
        Sink used to store quarantined results from node expectations.
    merge_cdc_options:
        Merge options to handle input DataFrames that are Change Data Capture
        (CDC). Only used when `merge` mode is selected.
    mode:
        Write mode.
        - overwrite: Overwrite existing data
        - append: Append contents of the dataframe to existing data
        - error: Throw and exception if data already exists
        - ignore: Silently ignore this operation if data already exists
        - complete: Overwrite for streaming dataframes
        - merge: Append, update and optionally delete records. Requires
        cdc specification.
    write_options:
        Other options passed to `spark.write.options`
    """

    is_quarantine: bool = False
    is_primary: bool = True
    checkpoint_location: str = None
    merge_cdc_options: DataSinkMergeCDCOptions = None  # TODO: Review parameter name
    mode: Union[
        Literal[
            "OVERWRITE", "APPEND", "IGNORE", "ERROR", "COMPLETE", "UPDATE", "MERGE"
        ],
        None,
    ] = None
    write_options: dict[str, str] = {}

    @model_validator(mode="after")
    def merge_has_options(self) -> Any:
        if self.mode == "MERGE":
            if self.merge_cdc_options is None:
                raise ValueError(
                    "If 'MERGE' `mode` is selected, `merge_cdc_options` must be specified."
                )
            else:
                self.merge_cdc_options._parent = self

        return self

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def _id(self):
        return str(self)

    @property
    def _uuid(self) -> str:
        hash_object = hashlib.sha1(self._id.encode())
        hash_digest = hash_object.hexdigest()
        return str(uuid.UUID(hash_digest[:32]))

    @property
    def _checkpoint_location(self) -> Path:
        if self.checkpoint_location:
            return Path(self.checkpoint_location)

        node = self.parent_pipeline_node

        if node and node._root_path:
            for i, s in enumerate(node.all_sinks):
                if s == self:
                    return node._root_path / "checkpoints" / f"sink-{self._uuid}"

        return None

    # ----------------------------------------------------------------------- #
    # CDC                                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def is_cdc(self) -> bool:
        return self.merge_cdc_options is not None

    @property
    def dlt_apply_changes_kwargs(self) -> dict[str, str]:
        """Keyword arguments for dlt.apply_changes function"""

        from laktory.models.datasinks.tabledatasink import TableDataSink

        if not isinstance(self, TableDataSink):
            raise ValueError("DLT only supports `TableDataSink` class")

        cdc = self.merge_cdc_options
        return {
            "apply_as_deletes": cdc.delete_where,
            # "apply_as_truncates": ,  # NOT SUPPORTED
            "column_list": cdc.include_columns,
            "except_column_list": cdc.exclude_columns,
            "ignore_null_updates": cdc.ignore_null_updates,
            "keys": cdc.primary_keys,
            "sequence_by": cdc.order_by,
            # "source": self.source.table_name,  # TO SET EXTERNALLY
            "stored_as_scd_type": cdc.scd_type,
            "target": self.table_name,
            # "track_history_column_list": cdc.track_history_columns,  # NOT SUPPORTED
            # "track_history_except_column_list": cdc.track_history_except_columns,  # NOT SUPPORTED
        }

    # ----------------------------------------------------------------------- #
    # Writers                                                                 #
    # ----------------------------------------------------------------------- #

    def write(
        self,
        df: AnyDataFrame = None,
        mode: str = None,
        spark=None,
        view_definition: str = None,
    ) -> None:
        """
        Write dataframe into sink.

        Parameters
        ----------
        df:
            Input dataframe.
        mode:
            Write mode overwrite of the sink default mode.
        spark:
            Spark Session for creating a view
        view_definition:
            View definition. Overwrites view definition defined in the sink.

        Returns
        -------
        """
        _view_definition = view_definition
        if _view_definition is None:
            _view_definition = getattr(self, "parsed_view_definition", None)
            if _view_definition:
                _view_definition = _view_definition.parsed_expr(view=True)

        if _view_definition:
            if self.df_backend == "SPARK":
                if spark is None:
                    raise ValueError(
                        "Spark session must be provided for creating a view."
                    )
                self._write_spark_view(view_definition=_view_definition, spark=spark)
            else:
                raise ValueError(
                    f"'{self.df_backend}' DataFrame backend is not supported for creating views"
                )

            logger.info("View created.")
            return

        if mode is None:
            mode = self.mode

        if is_spark_dataframe(df):
            self._write_spark(df=df, mode=mode)
        elif is_polars_dataframe(df=df):
            self._write_polars(df, mode=mode)
        else:
            raise ValueError(f"DataFrame type '{type(df)}' not supported")

        logger.info("Write completed.")

    def _write_spark(self, df: SparkDataFrame, mode: str = mode) -> None:
        raise NotImplementedError("Not implemented for Spark dataframe")

    def _write_spark_view(self, view_definition: str, spark) -> None:
        raise NotImplementedError(
            f"View creation with spark is not implemented for type '{type(self)}'"
        )

    def _write_polars(self, df: PolarsLazyFrame, mode=mode) -> None:
        raise NotImplementedError("Not implemented for Polars dataframe")

    # ----------------------------------------------------------------------- #
    # Purge                                                                   #
    # ----------------------------------------------------------------------- #

    def _purge_checkpoint(self, spark=None):
        if self._checkpoint_location:
            if os.path.exists(self._checkpoint_location):
                logger.info(
                    f"Deleting checkpoint at {self._checkpoint_location}",
                )
                shutil.rmtree(self._checkpoint_location)

            if spark is None:
                return

            try:
                from pyspark.dbutils import DBUtils
            except ModuleNotFoundError:
                return

            dbutils = DBUtils(spark)

            _path = self._checkpoint_location.as_posix()
            try:
                dbutils.fs.ls(
                    _path
                )  # TODO: Figure out why this does not work with databricks connect
                logger.info(
                    f"Deleting checkpoint at dbfs {_path}",
                )
                dbutils.fs.rm(_path, True)

            except Exception as e:
                if "java.io.FileNotFoundException" in str(e):
                    pass
                elif "databricks.sdk.errors.platform.ResourceDoesNotExist" in str(
                    type(e)
                ):
                    pass
                elif "databricks.sdk.errors.platform.InvalidParameterValue" in str(
                    type(e)
                ):
                    # TODO: Figure out why this is happening. It seems that the databricks SDK
                    #       modify the path before sending to REST API.
                    logger.warn(f"dbutils could not delete checkpoint {_path}: {e}")
                else:
                    raise e

    def purge(self):
        """
        Delete sink data and checkpoints
        """
        raise NotImplementedError()

    # ----------------------------------------------------------------------- #
    # Data Sources                                                            #
    # ----------------------------------------------------------------------- #

    def as_source(self, as_stream=None):
        raise NotImplementedError()

    def read(self, spark=None, as_stream=None):
        """
        Read dataframe from sink.

        Parameters
        ----------
        spark:
            Spark Session
        as_stream:
            If `True`, dataframe read as stream.

        Returns
        -------
        """
        return self.as_source(as_stream=as_stream).read(spark=spark)



from typing import Union

from laktory.models.basemodel import BaseModel
from laktory.models.resources.databricks.accesscontrol import AccessControl
from laktory.models.resources.databricks.permissions import Permissions
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class MLflowExperiment(BaseModel, PulumiResource, TerraformResource):
    """
    MLflow Experiment

    Attributes
    ----------
    access_controls:
        MLflow Experiment access controls
    artifact_location:
        Path to dbfs:/ or s3:// artifact location of the MLflow experiment.
    creation_time:
    description:
        The description of the MLflow experiment.
    experiment_id:
    last_update_time:
    lifecycle_stage:
    name:
        Name of MLflow experiment. It must be an absolute path within the
        Databricks workspace, e.g. `/Users/<some-username>/my-experiment`.
        For more information about changes to experiment naming conventions,
        see [mlflow docs](https://docs.databricks.com/applications/mlflow/experiments.html#experiment-migration).


    Examples
    --------
    ```py
    from laktory import models

    exp = models.resources.databricks.MLflowExperiment(
        name="/.laktory/Sample",
        artifact_location="dbfs:/tmp/my-experiment",
        description="My MLflow experiment description",
        access_controls=[
            {
                "group_name": "account users",
                "permission_level": "CAN_MANAGE",
            }
        ],
    )
    ```
    """

    access_controls: list[AccessControl] = []
    artifact_location: str = None
    creation_time: int = None
    description: str = None
    experiment_id: str = None
    last_update_time: int = None
    lifecycle_stage: str = None
    name: str

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        return self.name.replace("/", "_").replace("\\", "_").replace(" ", "_")

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - permissions
        """
        resources = []
        if self.access_controls:
            resources += [
                Permissions(
                    resource_name=f"permissions-{self.resource_name}",
                    access_controls=self.access_controls,
                    experiment_id=f"${{resources.{self.resource_name}.id}}",
                )
            ]

        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:MlflowExperiment"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["access_controls"]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_mlflow_experiment"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



from datetime import date
from datetime import datetime
from datetime import timezone
from typing import Union

from planck import units


def unix_timestamp(
    dt: Union[str, int, float, datetime, date] = None,
    unit: str = "s",
    as_int: bool = False,
) -> float:
    """
    Convert a datetime object into a unix timestamp float. If `None` is provided,
    current UTC timestamp is returned.

    Parameters
    ----------
    dt:
        Datetime object

    Returns
    -------
    output:
        Unix timestamp

    Examples
    --------
    ```py
    from datetime import datetime

    from laktory.datetime import unix_timestamp

    ts = unix_timestamp("2020-01-01T01:00:00")
    print(ts)
    # > 1577840400.0

    ts = unix_timestamp(datetime(2020, 1, 1, 1, 0, 0))
    print(ts)
    # > 1577840400.0
    ```
    """

    numeric_types = [int, float]
    np_installed = False
    np = None
    try:
        import numpy as np

        np_installed = True
        numeric_types += [np.float64, np.int64]
    except ModuleNotFoundError:
        pass

    if dt is None:
        dt = datetime.utcnow()
    elif type(dt) in numeric_types:
        dt = dt
    elif type(dt) == date:  # noqa: E721
        dt = datetime.combine(dt, datetime.min.time())
    elif isinstance(dt, str):
        try:
            dt = datetime.fromisoformat(dt)
        except ValueError:
            try:
                from dateutil import parser

                dt = parser.parse(dt)
            except ModuleNotFoundError:
                raise ValueError(
                    f"String '{dt}' is not a valid ISO 8601 datetime format. Install `dateutil` to support other formats."
                )
    elif np_installed and isinstance(dt, np.datetime64):
        dt = dt.astype(datetime)
        if isinstance(dt, int):
            dt = dt * units["ns"]["s"]

    if isinstance(dt, datetime) or isinstance(dt, date):
        # if pd.isnull(dt):
        #     dt = np.nan
        # else:
        if dt.tzinfo is None:
            dt = dt.replace(tzinfo=timezone.utc)
        dt = dt.timestamp()

    # Set units
    dt *= units["s"][unit]

    # Set type
    if as_int:
        dt = int(dt)

    return dt


def utc_datetime(
    unixtime: Union[datetime, date, str, float, int] = None,
) -> datetime:
    """
    Convert a unix timestamp into a datetime object. If `None` is provided,
    current UTC datetime object is returned.

    Parameters
    ----------
    unixtime:
        Unix timestamp

    Returns
    -------
    output:
        Datetime object

    Examples
    --------
    ```py
    from laktory.datetime import utc_datetime

    dt = utc_datetime("2020-01-01T01:00:00")
    print(dt)
    # > 2020-01-01 01:00:00

    dt = utc_datetime(1577840400)
    print(dt)
    # > 2020-01-01 01:00:00
    ```
    """

    np_installed = False
    np = None
    try:
        import numpy as np

        np_installed = True
    except ModuleNotFoundError:
        pass

    if not unixtime:
        dt = datetime.utcnow()
    elif isinstance(unixtime, datetime):
        dt = unixtime
    elif isinstance(unixtime, date):
        dt = datetime.combine(unixtime, datetime.min.time())
    elif isinstance(unixtime, str):
        dt = datetime.fromisoformat(unixtime)
    elif np_installed and np.isnan(unixtime):
        dt = None
    else:
        dt = datetime.fromtimestamp(unixtime, tz=timezone.utc).replace(tzinfo=None)

    return dt



from laktory.models.transformers.basechainnode import BaseChainNode
from laktory.models.transformers.polarschain import PolarsChain
from laktory.models.transformers.polarschainnode import PolarsChainNode
from laktory.models.transformers.polarschainnode import PolarsChainNodeFuncArg
from laktory.models.transformers.polarschainnode import PolarsChainNodeSQLExpr
from laktory.models.transformers.sparkchain import SparkChain
from laktory.models.transformers.sparkchainnode import SparkChainNode
from laktory.models.transformers.sparkchainnode import SparkChainNodeFuncArg
from laktory.models.transformers.sparkchainnode import SparkChainNodeSQLExpr



from pathlib import Path

import yaml

from laktory._parsers import _resolve_value

MERGE_KEY = "__merge_here"
VARIABLES_KEY = "variables"


# Register the constructors
class RecursiveLoader(yaml.SafeLoader):
    def __init__(self, stream, parent_loader=None):
        self.dirpath = Path("./")
        stream = self.preprocess_stream(stream)

        self.variables = []
        if parent_loader:
            self.variables = parent_loader.variables
        super().__init__(stream)

    def preprocess_stream(self, stream):
        """Reformat content to be YAML safe"""

        if hasattr(stream, "name"):
            self.dirpath = Path(stream.name).parent

        _lines = []
        for line in stream.readlines():
            if "${include." in line:
                raise ValueError(
                    "The `${include.}` syntax has been deprecated in laktory 0.6.0. Please use `!use`, `!update` and `!extend` tags instead."
                )
            _lines += [line.replace("<<:", MERGE_KEY + ":")]

        return "\n".join(_lines)

    @classmethod
    def load(cls, stream, parent_loader: "RecursiveLoader" = None):
        """
        Load yaml file with support for reference to external yaml and sql files using
        `!use`, `!extend` and `!update` tags.
        Path to external files can be defined using model or environment variables.

        Custom Tags
        -----------
        !use {filepath}:
            Directly inject the content of the file at `filepath`
        - !extend {filepath}:
            Extend the current list with the elements found in the file at `filepath`.
            Similar to python list.extend method.
        <<: !update {filepath}:
            Merge the current dictionary with the content of the dictionary defined at
            `filepath`. Similar to python dict.update method.

        Parameters
        ----------
        stream:
            file object structured as a yaml file
        parent_loader:
            Parent loader if file loader from another loader.

        Returns
        -------
        :
            Dict or list

        Examples
        --------
        ```yaml
        businesses:
          apple:
            symbol: aapl
            address: !use addresses.yaml
            <<: !update common.yaml
            emails:
              - jane.doe@apple.com
              - extend! emails.yaml
          amazon:
            symbol: amzn
            address: !use addresses.yaml
            <<: update! common.yaml
            emails:
              - john.doe@amazon.com
              - extend! emails.yaml
        ```
        """
        loader = cls(stream, parent_loader)
        try:
            return loader.get_single_data()
        finally:
            loader.dispose()

    @staticmethod
    def get_path(loader, node):
        filepath = Path(loader.construct_scalar(node))

        # set absolute path
        if not Path(filepath).is_absolute():
            filepath = loader.dirpath / filepath
        filepath = str(filepath)

        # resolve variables
        if "${vars." in filepath:
            variables = {}
            for _vars in loader.variables:
                variables.update(_vars)
            filepath = _resolve_value(str(filepath), variables)

            if "${vars." in filepath:
                raise ValueError(f"Some variables in {filepath} could not be resolved.")

        return filepath

    @staticmethod
    def inject_constructor(loader, node):
        """Inject content of another YAML file."""

        filepath = loader.get_path(loader, node)

        if filepath.endswith(".sql"):
            with open(filepath, "r", encoding="utf-8") as _fp:
                data = _fp.read()
            return '"' + data.replace('"', "'").replace("\n", "\\n") + '"'

        with open(filepath, "r") as f:
            return RecursiveLoader.load(f, parent_loader=loader)

    @staticmethod
    def merge_constructor(loader, node):
        """Merge content of another YAML file into the current dictionary."""

        filepath = loader.get_path(loader, node)
        with open(filepath, "r") as f:
            merge_data = RecursiveLoader.load(f, parent_loader=loader)

        if not isinstance(merge_data, dict):
            raise TypeError(
                f"Expected a dictionary in {filepath}, but got {type(merge_data).__name__}"
            )
        return merge_data

    @staticmethod
    def append_constructor(loader, node):
        """Append content of another YAML file to the current list."""

        filepath = loader.get_path(loader, node)
        with open(filepath, "r") as f:
            append_data = RecursiveLoader.load(f, parent_loader=loader)
        if not isinstance(append_data, list):
            raise TypeError(
                f"Expected a list in {filepath}, but got {type(append_data).__name__}"
            )
        return append_data

    # Custom mapping constructor to handle merging
    @staticmethod
    def custom_mapping_constructor(loader, node):
        """Custom handling for mappings to support !merge."""

        # read variables
        _vars = {}
        for key_node, value_node in node.value:
            key = loader.construct_object(key_node)
            if key == VARIABLES_KEY:
                _vars = loader.construct_object(value_node)
        loader.variables += [_vars]

        # read include and merge
        mapping = {}
        for key_node, value_node in node.value:
            key = loader.construct_object(key_node)
            value = loader.construct_object(value_node)
            if key == MERGE_KEY and isinstance(value, dict):
                # Merge the dictionary directly into the parent mapping
                mapping.update(value)
            else:
                mapping[key] = value

        # remove variables
        del loader.variables[-1]

        return mapping

    # Custom sequence constructor to handle appending
    @staticmethod
    def custom_sequence_constructor(loader, node):
        """Custom handling for sequences to support !append."""

        seq = []
        for child in node.value:
            if child.tag == "!extend":
                append_data = loader.construct_object(child)
                seq.extend(append_data)  # Flatten the appended list
            else:
                seq.append(loader.construct_object(child))
        return seq


RecursiveLoader.add_constructor("!use", RecursiveLoader.inject_constructor)
RecursiveLoader.add_constructor("!update", RecursiveLoader.merge_constructor)
RecursiveLoader.add_constructor("!extend", RecursiveLoader.append_constructor)
RecursiveLoader.add_constructor(
    yaml.resolver.BaseResolver.DEFAULT_SEQUENCE_TAG,
    RecursiveLoader.custom_sequence_constructor,
)
RecursiveLoader.add_constructor(
    yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
    RecursiveLoader.custom_mapping_constructor,
)



from functools import wraps

import pyspark.sql.functions as F

import laktory.spark.functions.logical as logical
import laktory.spark.functions.math as math
import laktory.spark.functions.string as string
import laktory.spark.functions.units as units

# For documentation
from laktory.spark.functions.logical import *  # noqa: F403
from laktory.spark.functions.math import *  # noqa: F403
from laktory.spark.functions.string import *  # noqa: F403
from laktory.spark.functions.units import *  # noqa: F403


class LaktoryFunctions:
    # ----------------------------------------------------------------------- #
    # Logical                                                                 #
    # ----------------------------------------------------------------------- #

    @staticmethod
    def compare(*args, **kwargs):
        return logical.compare(*args, **kwargs)

    # ----------------------------------------------------------------------- #
    # Math                                                                    #
    # ----------------------------------------------------------------------- #

    @staticmethod
    @wraps(math.roundp)
    def roundp(*args, **kwargs):
        return math.roundp(*args, **kwargs)

    # ----------------------------------------------------------------------- #
    # String                                                                  #
    # ----------------------------------------------------------------------- #

    @staticmethod
    @wraps(string.string_split)
    def string_split(*args, **kwargs):
        return string.string_split(*args, **kwargs)

    @staticmethod
    @wraps(string.uuid)
    def uuid(*args, **kwargs):
        return string.uuid(*args, **kwargs)

    # ----------------------------------------------------------------------- #
    # Units                                                                   #
    # ----------------------------------------------------------------------- #

    @staticmethod
    @wraps(units.convert_units)
    def convert_units(*args, **kwargs):
        return units.convert_units(*args, **kwargs)


F.laktory = LaktoryFunctions()



from typing import Literal
from typing import Union

from laktory.models.basemodel import BaseModel
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource

# class VectorSearchEndpointLookup(ResourceLookup):
#     """
#     Attributes
#     ----------
#     id:
#         The ID of the Vector Search Endpoint warehouse.
#     """
#
#     id: str = Field(serialization_alias="id")


class VectorSearchEndpoint(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Warehouse

    Attributes
    ----------
    endpoint_type:
        Type of Vector Search Endpoint.
        Currently only accepting single value: STANDARD
    name:
        Name of the Vector Search Endpoint to create.

    Examples
    --------
    ```py
    from laktory import models

    endpoint = models.resources.databricks.VectorSearchEndpoint(
        endpoint_type="STANDARD",
        name="default",
    )
    ```
    """

    endpoint_type: Literal["STANDARD"] = "STANDARD"  # required
    name: str = None

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:VectorSearchEndpoint"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["access_controls"]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_vector_search_endpoint"

    # @property
    # def terraform_resource_lookup_type(self) -> str:
    #     return "databricks_sql_warehouse"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



import polars.datatypes as T

DATATYPES_MAP = {
    "binary": T.Binary,
    "byte": T.Int8,
    "int8": T.Int8,
    "tinyint": T.Int8,
    "short": T.Int16,
    "int16": T.Int16,
    "smallint": T.Int16,
    "int": T.Int32,
    "int32": T.Int32,
    "long": T.Int64,
    "int64": T.Int64,
    "bigint": T.Int64,
    "float": T.Float32,
    "float32": T.Float32,
    "double": T.Float64,
    "float64": T.Float64,
    "boolean": T.Boolean,
    "string": T.Utf8,
    "utf8": T.Utf8,
    "date": T.Date,
    "timestamp": T.Datetime,
    "datetime": T.Datetime,
    # "struct<>": T.Struct,
    # "struct": T.Struct,
    # "array": T.List,
    # "list": T.List,
}



from typing import Any
from typing import Literal
from typing import Union

from pydantic import Field
from pydantic import model_validator

from laktory._logger import get_logger
from laktory.models.basemodel import BaseModel
from laktory.models.pipeline.pipelinechild import PipelineChild
from laktory.polars import PolarsDataFrame
from laktory.polars import is_polars_dataframe
from laktory.spark import SparkDataFrame
from laktory.spark import is_spark_dataframe
from laktory.typing import AnyDataFrame

logger = get_logger(__name__)


class DataFrameSample(BaseModel):
    fraction: float
    seed: Union[int, None] = None


class Watermark(BaseModel):
    """
    Definition of a spark structured streaming watermark for joining data
    streams.

    Attributes
    ----------
    column:
        Event time column name
    threshold:
        How late, expressed in seconds, the data is expected to be with
        respect to event time.

    References
    ----------
    https://spark.apache.org/docs/latest/structured-streaming-programming-guide.html#handling-late-data-and-watermarking
    """

    column: str
    threshold: str


class BaseDataSource(BaseModel, PipelineChild):
    """
    Base class for building data source

    Attributes
    ----------
    as_stream:
        If `True`source is read as a streaming DataFrame.
    broadcast:
        If `True` DataFrame is broadcasted
    dataframe_backend:
        Type of dataframe
    drops:
        List of columns to drop
    filter:
        SQL expression used to select specific rows from the source table
    renames:
        Mapping between the source table column names and new column names
    selects:
        Columns to select from the source table. Can be specified as a list
        or as a dictionary to rename the source columns
    watermark
        Spark structured streaming watermark specifications
    """

    as_stream: bool = False
    broadcast: Union[bool, None] = False
    dataframe_backend: Literal["SPARK", "POLARS"] = None
    drops: Union[list, None] = None
    filter: Union[str, None] = None
    limit: Union[int, None] = None
    mock_df: Any = Field(default=None, exclude=True)
    renames: Union[dict[str, str], None] = None
    sample: Union[DataFrameSample, None] = None
    selects: Union[list[str], dict[str, str], None] = None
    watermark: Union[Watermark, None] = None

    @model_validator(mode="after")
    def options(self) -> Any:
        with self.validate_assignment_disabled():
            # Overwrite Dataframe type if mock dataframe is provided
            if is_spark_dataframe(self.mock_df):
                self.dataframe_backend = "SPARK"
            elif is_polars_dataframe(self.mock_df):
                self.dataframe_backend = "POLARS"

            if self.df_backend == "SPARK":
                pass
            elif self.df_backend == "POLARS":
                if self.as_stream:
                    raise ValueError("Polars DataFrames don't support streaming read.")
                if self.watermark:
                    raise ValueError("Polars DataFrames don't support watermarking.")
                if self.broadcast:
                    raise ValueError("Polars DataFrames don't support broadcasting.")

        return self

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def _id(self):
        return str(self.df)

    @property
    def is_orchestrator_dlt(self) -> bool:
        """If `True`, data source is used in the context of a DLT pipeline"""

        pl = self.parent_pipeline

        if pl is None:
            return False

        return pl.is_orchestrator_dlt

    # ----------------------------------------------------------------------- #
    # Readers                                                                 #
    # ----------------------------------------------------------------------- #

    def read(self, spark=None) -> AnyDataFrame:
        """
        Read data with options specified in attributes.

        Parameters
        ----------
        spark:
            Spark context

        Returns
        -------
        : DataFrame
            Resulting dataframe
        """
        if self.mock_df is not None:
            df = self.mock_df
        elif self.df_backend == "SPARK":
            df = self._read_spark(spark=spark)
        elif self.df_backend == "POLARS":
            df = self._read_polars()
        else:
            raise ValueError(f"DataFrame type '{self.df_backend}' is not supported.")

        if is_spark_dataframe(df):
            df = self._post_read_spark(df)
        elif is_polars_dataframe(df):
            df = self._post_read_polars(df)

        logger.info("Read completed.")

        return df

    def _read_spark(self, spark) -> SparkDataFrame:
        raise NotImplementedError()

    def _read_polars(self) -> PolarsDataFrame:
        raise NotImplementedError()

    def _post_read_spark(self, df: SparkDataFrame) -> SparkDataFrame:
        import pyspark.sql.functions as F

        # Apply filter
        if self.filter:
            df = df.filter(self.filter)

        # Apply drops
        if self.drops:
            df = df.drop(*self.drops)

        # Columns
        cols = []
        if self.selects:
            if isinstance(self.selects, list):
                cols += [F.col(c) for c in self.selects]
            elif isinstance(self.selects, dict):
                cols += [F.col(k).alias(v) for k, v in self.selects.items()]
            df = df.select(cols)

        # Renames
        if self.renames:
            for old_name, new_name in self.renames.items():
                df = df.withColumnRenamed(old_name, new_name)

        # Apply Watermark
        if self.watermark:
            df = df.withWatermark(
                self.watermark.column,
                self.watermark.threshold,
            )

        # Broadcast
        if self.broadcast:
            df = F.broadcast(df)

        # Sample
        if self.sample:
            df = df.sample(fraction=self.sample.fraction, seed=self.sample.seed)

        # Limit
        if self.limit:
            df = df.limit(self.limit)

        return df

    def _post_read_polars(self, df: PolarsDataFrame) -> PolarsDataFrame:
        import polars as pl

        from laktory.polars.expressions.sql import _parse_token

        # Apply filter
        if self.filter:
            df = df.filter(pl.Expr.laktory.sql_expr(self.filter))

        # Columns
        cols = []
        if self.selects:
            if isinstance(self.selects, list):
                cols += [_parse_token(c) for c in self.selects]
            elif isinstance(self.selects, dict):
                cols += [_parse_token(k).alias(v) for k, v in self.selects.items()]
            df = df.select(cols)

        # Apply drops
        if self.drops:
            df = df.drop(*self.drops)

        # Renames
        if self.renames:
            df = df.rename(self.renames)

        # Apply Watermark
        if self.watermark:
            raise NotImplementedError(
                "Watermarking not supported with POLARS dataframe"
            )

        # Broadcast
        if self.broadcast:
            raise NotImplementedError(
                "Broadcasting not supported with POLARS dataframe"
            )

        # Sample
        if self.sample:
            raise NotImplementedError(
                "Broadcasting not supported with POLARS lazyframe"
            )
            # df = df.sample(fraction=self.sample.fraction, seed=self.sample.seed)

        # Limit
        if self.limit:
            df = df.limit(self.limit)

        return df



from typing import Union

from laktory._settings import settings
from laktory.models.datasinks.tabledatasink import TableDataSink
from laktory.models.pipeline.orchestrators.pipelineconfigworkspacefile import (
    PipelineConfigWorkspaceFile,
)
from laktory.models.pipeline.orchestrators.pipelinerequirementsworkspacefile import (
    PipelineRequirementsWorkspaceFile,
)
from laktory.models.pipeline.pipelinechild import PipelineChild
from laktory.models.resources.databricks.dltpipeline import DLTPipeline
from laktory.models.resources.pulumiresource import PulumiResource


class DatabricksDLTOrchestrator(DLTPipeline, PipelineChild):
    """
    Databricks DLT used as an orchestrator to execute a Laktory pipeline.

    DLT orchestrator does not support pipeline nodes with views (as opposed to
    materialized tables). Also, it does not support writing to multiple
    schemas within the same pipeline.

    Selecting this orchestrator requires to add the supporting
    [notebook](https://github.com/okube-ai/laktory/blob/main/laktory/resources/quickstart-stacks/workflows/notebooks/dlt/dlt_laktory_pl.py)
    to the stack.

    Attributes
    ----------
    config_file:
        Pipeline configuration (json) file deployed to the workspace and used
        by the job to read and execute the pipeline.
    requirements_file:
        Pipeline requirements (json) file deployed to the workspace and used
        by the job to install the required python dependencies.
    """

    config_file: PipelineConfigWorkspaceFile = PipelineConfigWorkspaceFile()
    requirements_file: PipelineRequirementsWorkspaceFile = (
        PipelineRequirementsWorkspaceFile()
    )

    # ----------------------------------------------------------------------- #
    # Update DLT                                                              #
    # ----------------------------------------------------------------------- #

    def update_from_parent(self):
        pl = self.parent_pipeline

        for node in pl.nodes:
            if node.is_view:
                raise ValueError(
                    f"Node '{node.name}' of pipeline '{pl.name}' is a view which is not supported with DLT orchestrator."
                )

        for n in pl.nodes:
            for s in n.all_sinks:
                if isinstance(s, TableDataSink):
                    s.catalog_name = self.catalog
                    s.schema_name = self.target

        # Configuration
        self.configuration["pipeline_name"] = pl.name
        self.configuration["workspace_laktory_root"] = settings.workspace_laktory_root

        # Config file
        self.config_file.update_from_parent()

        # Requirements file
        self.requirements_file.update_from_parent()

    # ----------------------------------------------------------------------- #
    # Children                                                                #
    # ----------------------------------------------------------------------- #

    @property
    def child_attribute_names(self):
        return ["config_file", "requirements_file"]

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_type_id(self) -> str:
        return "dlt-pipeline"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        excludes = super().pulumi_excludes
        excludes["config_file"] = True
        excludes["requirements_file"] = True
        return excludes

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - configuration workspace file
        - configuration workspace file permissions
        - requirements workspace file
        - requirements workspace file permissions
        """

        resources = super().additional_core_resources
        resources += [self.config_file]
        resources[-1].write_source()
        resources += [self.requirements_file]
        resources[-1].write_source()

        return resources



from typing import Union

from pydantic import Field

from laktory.models.basemodel import BaseModel
from laktory.models.resources.baseresource import ResourceLookup
from laktory.models.resources.databricks.mwspermissionassignment import (
    MwsPermissionAssignment,
)
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class GroupLookup(ResourceLookup):
    """
    Attributes
    ----------
    id:
        Id of the group. Only supported when using Pulumi backend.
    display_name:
        Display name of the group. Only support when using Terraform backend
    """

    id: str = None
    display_name: str = None


class Group(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks group

    Attributes
    ----------
    allow_cluster_create:
        When `True`, the group is allowed to have cluster create permissions
    display_name:
        Display name for the group.
    lookup_existing:
        Specifications for looking up existing resource. Other attributes will
        be ignored.
    workspace_access
        When `True`, the group is allowed to have workspace access

    Examples
    --------
    ```py
    from laktory import models

    d = models.resources.databricks.Group(display_name="role-engineers")
    ```
    """

    allow_cluster_create: bool = False
    display_name: str
    lookup_existing: GroupLookup = Field(None, exclude=True)
    workspace_access: bool = None
    workspace_permission_assignments: list[MwsPermissionAssignment] = None

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        """display name"""
        return self.display_name

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:Group"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["workspace_permission_assignments"]

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - workspace permission assignments
        """
        resources = []
        if self.workspace_permission_assignments:
            for a in self.workspace_permission_assignments:
                if a.principal_id is None:
                    a.principal_id = f"${{resources.{self.resource_name}.id}}"
                resources += [a]
        return resources

    # TODO:
    # if group.id is None:
    #     self.group = databricks.Group(name, opts=opts, **group.model_pulumi_dump())
    #     group.id = self.group.id
    # else:
    #     self.group = databricks.Group.get(name, id=group.id)

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_group"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



from typing import Any

from pydantic import model_validator

from laktory.models.basemodel import BaseModel


class AccessControl(BaseModel):
    """
    Databricks Access Control

    Access Control generally applicable to objects like cluster, notebook,
    pipeline, etc. For providing access to securable data, refer to
    `models.Grant` instead.

    Attributes
    ----------
    group_name:
        Name of the group to assign the permission to.
    permission_level:
        Name of the permission to assign
    service_principal_name:
        Name of the service principal to assign the permission to.
    user_name
        Name of the user to assign the permission to.

    Examples
    --------
    ```py
    from laktory import models

    p = models.resources.databricks.AccessControl(
        group_name="role-engineers", permission_level="READ"
    )
    ```
    """

    group_name: str = None
    permission_level: str
    service_principal_name: str = None
    user_name: str = None

    @model_validator(mode="after")
    def single_input(self) -> Any:
        count = 0
        if self.user_name:
            count += 1
        if self.group_name:
            count += 1
        if self.service_principal_name:
            count += 1

        if count == 0:
            raise ValueError(
                "At least one of `user_name`, `group_name`, `service_principal_name` must be specified."
            )
        if count > 1:
            raise ValueError(
                "Ony one of `user_name`, `group_name`, `service_principal_name` must be specified."
            )

        return self



import polars as pl

from laktory.datetime import utc_datetime

__all__ = [
    "current_timestamp",
]


# --------------------------------------------------------------------------- #
# compare                                                                     #
# --------------------------------------------------------------------------- #


def current_timestamp() -> pl.Expr:
    """
    Returns the current timestamp at the start of expression evaluation as a
    Datetime column.

    Returns
    -------
    :
        Current timestamp

    Examples
    --------
    ```py
    import polars as pl

    import laktory  # noqa: F401

    df = pl.DataFrame({"x": [0.45, 0.55]})
    df = df.with_columns(tstamp=pl.expr.laktory.current_timestamp())
    ```
    """

    return pl.lit(utc_datetime())



import polars as pl


def schema_flat(df: pl.DataFrame) -> list[str]:
    """
    Returns a flattened list of columns

    Parameters
    ----------
    df:
        Input DataFrame

    Returns
    -------
    :
        List of columns

    Examples
    --------
    ```py
    import polars as pl

    import laktory  # noqa: F401

    df = pl.DataFrame(
        {
            "indexx": [1, 2, 3],
            "stock": [
                {"symbol": "AAPL", "name": "Apple"},
                {"symbol": "MSFT", "name": "Microsoft"},
                {"symbol": "GOOGL", "name": "Google"},
            ],
            "prices": [
                [{"open": 1, "close": 2}, {"open": 1, "close": 2}],
                [{"open": 1, "close": 2}, {"open": 1, "close": 2}],
                [{"open": 1, "close": 2}, {"open": 1, "close": 2}],
            ],
        }
    )

    print(df.laktory.schema_flat())
    '''
    [
        'indexx',
        'stock',
        'stock.symbol',
        'stock.name',
        'prices',
        'prices[*].open',
        'prices[*].close',
    ]
    '''
    ```
    """

    def get_fields(schema):
        field_names = []
        for f_name, f_type in schema.items():
            if isinstance(f_type, pl.Struct):
                _field_names = get_fields(dict(f_type))
                field_names += [f_name]
                field_names += [f"{f_name}.{v}" for v in _field_names]

            elif isinstance(f_type, pl.List):
                field_names += [f_name]
                if isinstance(f_type.inner, pl.Struct):
                    _field_names = get_fields(dict(f_type.inner))
                    field_names += [f"{f_name}[*].{v}" for v in _field_names]

            else:
                field_names += [f_name]

        return field_names

    return get_fields(df.schema)



from typing import Literal

from pydantic import ConfigDict

from laktory.models.basemodel import BaseModel
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class MlflowWebhookHttpUrlSpec(BaseModel):
    """
    MLflow webhook URL Specifications

    Attributes
    ----------
    url:
        External HTTPS URL called on event trigger (by using a POST request). Structure of payload depends on the event
        type, refer to [documentation](https://docs.databricks.com/applications/mlflow/model-registry-webhooks.html)
        for more details.
    authorization:
        Value of the authorization header that should be sent in the request sent by the wehbook. It should be of the
        form `<auth type> <credentials>`, e.g. `Bearer <access_token>`. If set to an empty string, no authorization
        header will be included in the request.
    enable_ssl_verification:
        Enable/disable SSL certificate validation. Default is `true`. For self-signed certificates, this field must be
        `false` AND the destination server must disable certificate validation as well. For security purposes, it is
        encouraged to perform secret validation with the HMAC-encoded portion of the payload and acknowledge the risk
        associated with disabling hostname validation whereby it becomes more likely that requests can be maliciously
        routed to an unintended host.
    secret:
        Shared secret required for HMAC encoding payload. The HMAC-encoded payload will be sent in the header as
        `X-Databricks-Signature: encoded_payload`.

    """

    url: str
    authorization: str = None
    enable_ssl_verification: bool = None
    secret: str = None


class MlflowWebhookJobSpec(BaseModel):
    """
    MLflow webhook Job Specifications

    Attributes
    ----------
    access_token:
        The personal access token used to authorize webhook's job runs.
    job_id:
        ID of the Databricks job that the webhook runs.
    workspace_url:
        URL of the workspace containing the job that this webhook runs. If not specified, the job’s workspace URL is
        assumed to be the same as the workspace where the webhook is created.
    """

    access_token: str
    job_id: str
    workspace_url: str = None


class MLflowWebhook(BaseModel, PulumiResource, TerraformResource):
    """
    MLflow Model

    Attributes
    ----------
    events:
        The list of events that will trigger execution of Databricks job or POSTing to an URL, for example,
        `MODEL_VERSION_CREATED`, `MODEL_VERSION_TRANSITIONED_STAGE`, `TRANSITION_REQUEST_CREATED`, etc.
        Refer to the (Webhooks API documentation)[https://docs.databricks.com/dev-tools/api/latest/mlflow.html#operation/create-registry-webhook]
        for a full list of supported events.
    description:
        Optional description of the MLflow webhook.
    http_url_spec:
        URL Specifications
    job_spec:
        Job Specifications
    model_name:
        Name of MLflow model for which webhook will be created. If the model name is not specified, a registry-wide
        webhook is created that listens for the specified events across all versions of all registered models.
    status:
        Optional status of webhook. Default is `ACTIVE`

    Examples
    --------
    ```py
    from laktory import models

    mlwebhook = models.resources.databricks.MLflowWebhook(
        events=["TRANSITION_REQUEST_CREATED"],
        description="Databricks Job webhook trigger",
        status="ACTIVE",
        job_spec={
            "job_id": "some_id",
            "workspace_url": "some_url",
            "access_token": "some_token",
        },
    )
    ```
    """

    model_config = ConfigDict(protected_namespaces=())
    events: list[str]
    description: str = None
    http_url_spec: MlflowWebhookHttpUrlSpec = None
    job_spec: MlflowWebhookJobSpec = None
    model_name: str = None
    status: Literal["ACTIVE", "TEST_MODE", "DISABLED"] = None

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        key = ""
        if self.model_name:
            key += self.model_name + "-"
        for e in self.events:
            key += e + "-"

        return key.replace(" ", "_")

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    # @property
    # def pulumi_renames(self) -> dict[str, str]:
    #     return {"modelname": "model_name"}

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:MlflowWebhook"

    # @property
    # def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
    #     return ["access_controls"]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_mlflow_webhook"

    # @property
    # def terraform_renames(self) -> dict[str, str]:
    #     return self.pulumi_renames

    # @property
    # def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
    #     return self.pulumi_excludes



from pathlib import Path
from typing import Any
from typing import Union

from pydantic import model_validator

from laktory._settings import settings
from laktory.models.basemodel import BaseModel
from laktory.models.resources.databricks.accesscontrol import AccessControl
from laktory.models.resources.databricks.permissions import Permissions
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class AlertConditionThresholdValue(BaseModel):
    """
    Attributes
    ----------
    bool_value:
        boolean value (`true` or `false`) to compare against boolean results.
    double_value:
        double value to compare against integer and double results.
    string_value:
        string value to compare against string results.
    """

    bool_value: bool = None
    double_value: float = None
    string_value: str = None


class AlertConditionThreshold(BaseModel):
    """
    Attributes
    ----------
    value:
        Actual value used in comparison (one of the attributes is required)
    """

    value: AlertConditionThresholdValue


class AlertConditionOperandColumn(BaseModel):
    """
    Attributes
    ----------
    name:
        Name of the column
    """

    name: str


class AlertConditionOperand(BaseModel):
    """
    Attributes
    ----------
    column:
        Block describing the column from the query result to use for comparison
        in alert evaluation
    """

    column: AlertConditionOperandColumn


class AlertCondition(BaseModel):
    """
    Alert Conditions

    Attributes
    ----------
    op:
        Operator used to compare in alert evaluation. (Enum: `GREATER_THAN`,
        `GREATER_THAN_OR_EQUAL`, `LESS_THAN`, `LESS_THAN_OR_EQUAL`, `EQUAL`,
        `NOT_EQUAL`, `IS_NULL`)
    operand:
        Name of the column from the query result to use for comparison in alert
        evaluation.
    empty_result_state:
        Alert state if the result is empty (`UNKNOWN`, `OK`, `TRIGGERED`)
    threshold:
        Threshold value used for comparison in alert evaluation:
    """

    op: str
    operand: AlertConditionOperand
    empty_result_state: str = None
    threshold: AlertConditionThreshold


class Alert(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks SQL Query

    Attributes
    ----------
    access_controls:
        SQL Alert access controls
    condition:
        Trigger conditions of the alert.
    dirpath:
        Workspace directory inside rootpath in which the alert is deployed.
        Used only if `parent_path` is not specified.
    display_name:
        Name of the alert.
    query_id:
        ID of the query evaluated by the alert. Mutually exclusive with `query`.
    custom_body:
        Custom body of alert notification, if it exists. See [Alerts API reference](https://docs.databricks.com/sql/user/alerts/index.html)
        for custom templating instructions.
    custom_subject:
        Custom subject of alert notification, if it exists. This includes email subject, Slack notification header, etc.
        See [Alerts API reference](https://docs.databricks.com/sql/user/alerts/index.html) for custom templating
        instructions.
    name_prefix:
        Prefix added to the alert display name
    name_suffix:
        Suffix added to the alert display name
    notify_on_ok:
        Whether to notify alert subscribers when alert returns back to normal.
    owner_user_name:
        Alert owner's username.
    parent_path:
        The path to a workspace folder containing the alert. Set to `None`
        to use user's home folder. Overwrite `rootpath` and `dirpath`. If
        changed, the alert will be recreated.
    rootpath:
        Root directory to which all alerts are deployed to. Can also be
        configured by settings LAKTORY_WORKSPACE_LAKTORY_ROOT environment
        variable. Default is `/.laktory/`. Used only if `parent_path` is not
        specified.
    seconds_to_retrigger:
        Number of seconds an alert must wait after being triggered to rearm
        itself. After rearming, it can be triggered again. If 0 or not
        specified, the alert will not be triggered again.

    Examples
    --------
    ```py
    from laktory import models

    alert = models.resources.databricks.Alert(
        display_name="My Alert",
        parent_path="/alerts",
        query_id="1",
        condition={
            "op": "GREATER_THAN",
            "operand": {"column": {"name": "value"}},
            "threshold": {"value": {"double_value": 42.0}},
        },
    )
    ```
    """

    access_controls: list[AccessControl] = []
    condition: AlertCondition
    custom_body: str = None
    custom_subject: str = None
    dirpath: str = None
    display_name: str
    name_prefix: str = None
    name_suffix: str = None
    notify_on_ok: bool = None
    owner_user_name: str = None
    parent_path: str = None
    query_id: str = None
    rootpath: str = None
    seconds_to_retrigger: int = None

    @model_validator(mode="after")
    def set_paths(self) -> Any:
        # Parent Path explicitly set
        if "parent_path" in self.model_fields_set:
            return self

        # root
        if self.rootpath is None:
            self.rootpath = settings.workspace_laktory_root

        # dir
        if self.dirpath is None:
            self.dirpath = ""
        if self.dirpath.startswith("/"):
            self.dirpath = self.dirpath[1:]

        # parent_path
        _path = Path(self.rootpath) / self.dirpath
        self.parent_path = _path.as_posix()

        return self

    @model_validator(mode="after")
    def update_name(self) -> Any:
        if self.name_prefix:
            self.display_name = self.name_prefix + self.display_name
            self.name_prefix = ""
        if self.name_suffix:
            self.display_name = self.display_name + self.name_suffix
            self.name_suffix = ""
        return self

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        return self.display_name.replace(" ", "_")

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - permissions
        """
        resources = []
        if self.access_controls:
            resources += [
                Permissions(
                    resource_name=f"permissions-{self.resource_name}",
                    access_controls=self.access_controls,
                    sql_alert_id=f"${{resources.{self.resource_name}.id}}",
                )
            ]

        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:Alert"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["access_controls", "dirpath", "rootpath", "name_prefix", "name_suffix"]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_alert"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



from typing import Union

from laktory._settings import settings
from laktory.models.pipeline.orchestrators.pipelineconfigworkspacefile import (
    PipelineConfigWorkspaceFile,
)
from laktory.models.pipeline.orchestrators.pipelinerequirementsworkspacefile import (
    PipelineRequirementsWorkspaceFile,
)
from laktory.models.pipeline.pipelinechild import PipelineChild
from laktory.models.resources.databricks.cluster import ClusterLibrary
from laktory.models.resources.databricks.cluster import ClusterLibraryPypi
from laktory.models.resources.databricks.job import Job
from laktory.models.resources.databricks.job import JobParameter
from laktory.models.resources.databricks.job import JobTask
from laktory.models.resources.pulumiresource import PulumiResource


class DatabricksJobOrchestrator(Job, PipelineChild):
    """
    Databricks job used as an orchestrator to execute a Laktory pipeline.

    Job orchestrator supports incremental workloads with Spark Structured
    Streaming, but it does not support continuous processing.

    Selecting this orchestrator requires to add the supporting
    [notebook](https://github.com/okube-ai/laktory/blob/main/laktory/resources/quickstart-stacks/workflows/notebooks/jobs/job_laktory_pl.py)
    to the stack.

    Attributes
    ----------
    notebook_path:
        Path for the notebook. If `None`, default path for laktory job notebooks is used.
    config_file:
        Pipeline configuration (json) file deployed to the workspace and used
        by the job to read and execute the pipeline.
    requirements_file:
        Pipeline requirements (json) file deployed to the workspace and used
        by the job to install the required python dependencies.
    """

    notebook_path: Union[str, None] = None
    config_file: PipelineConfigWorkspaceFile = PipelineConfigWorkspaceFile()
    requirements_file: PipelineRequirementsWorkspaceFile = (
        PipelineRequirementsWorkspaceFile()
    )

    # ----------------------------------------------------------------------- #
    # Update Job                                                              #
    # ----------------------------------------------------------------------- #

    def update_from_parent(self):
        cluster_found = False
        for c in self.clusters:
            if c.name == "node-cluster":
                cluster_found = True
        if len(self.clusters) > 0 and not cluster_found:
            raise ValueError(
                "To use DATABRICKS_JOB orchestrator, a cluster named `node-cluster` must be defined in the databricks_job attribute."
            )

        pl = self.parent_pipeline

        self.parameters = [
            JobParameter(name="full_refresh", default="false"),
            JobParameter(name="pipeline_name", default=pl.name),
            JobParameter(
                name="install_dependencies", default=str(not cluster_found).lower()
            ),
        ]

        notebook_path = self.notebook_path
        if notebook_path is None:
            notebook_path = f"{settings.workspace_laktory_root}jobs/job_laktory_pl.py"

        self.tasks = []

        # Sorting Node Names to prevent job update trigger with Pulumi
        node_names = [node.name for node in pl.nodes]
        node_names.sort()
        for node_name in node_names:
            node = pl.nodes_dict[node_name]

            depends_on = []
            for edge in pl.dag.in_edges(node.name):
                depends_on += [{"task_key": "node-" + edge[0]}]

            libraries = []
            if cluster_found:
                libraries = [
                    ClusterLibrary(pypi=ClusterLibraryPypi(package=d))
                    for d in pl._dependencies
                ]

            self.tasks += [
                JobTask(
                    task_key="node-" + node.name,
                    notebook_task={
                        "base_parameters": {"node_name": node.name},
                        "notebook_path": notebook_path,
                    },
                    libraries=libraries,
                    depends_ons=depends_on,
                )
            ]
            self.sort_tasks(self.tasks)

            if cluster_found:
                self.tasks[-1].job_cluster_key = "node-cluster"

        # Config file
        self.config_file.update_from_parent()

        # Requirements file
        self.requirements_file.update_from_parent()

    # ----------------------------------------------------------------------- #
    # Children                                                                #
    # ----------------------------------------------------------------------- #

    @property
    def child_attribute_names(self):
        return ["config_file", "requirements_file"]

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_type_id(self) -> str:
        return "job"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return super().pulumi_excludes + [
            "notebook_path",
            "config_file",
            "requirements_file",
        ]

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - configuration workspace file
        - configuration workspace file permissions
        - requirements workspace file
        - requirements workspace file permissions
        """

        resources = super().additional_core_resources
        resources += [self.config_file]
        resources[-1].write_source()
        resources += [self.requirements_file]
        resources[-1].write_source()

        return resources



import laktory.cli._deploy
import laktory.cli._destroy
import laktory.cli._init
import laktory.cli._preview
import laktory.cli._quickstart
import laktory.cli._run
import laktory.cli._version
from laktory.cli._deploy import deploy
from laktory.cli._destroy import destroy
from laktory.cli._init import init
from laktory.cli._preview import preview
from laktory.cli._quickstart import quickstart
from laktory.cli._run import run
from laktory.cli._version import version
from laktory.cli.app import app
from laktory.cli.app import main



from .dataframesfactory import DataFramesFactory
from .monkeypatch import MonkeyPatch
from .paths import Paths
from .sparkfactory import SparkFactory
from .stackvalidator import StackValidator

sparkf = SparkFactory()
dff = DataFramesFactory(sparkf)



spark_installed = True
try:
    import pyspark
except ModuleNotFoundError:
    spark_installed = False

if spark_installed:
    from pyspark.sql.column import Column as SparkColumn
    from pyspark.sql.connect.dataframe import DataFrame as SparkConnectDataFrame
    from pyspark.sql.dataframe import DataFrame as SparkDataFrame
    from pyspark.sql.session import SparkSession

    import laktory.spark.connectsession
    import laktory.spark.dataframe
    import laktory.spark.functions
    import laktory.spark.session
    from laktory.spark.datatypes import DATATYPES_MAP

    def is_spark_dataframe(df):
        """Check if dataframe is Spark DataFrame or Spark Connect DataFrame"""

        if isinstance(df, SparkConnectDataFrame):
            return True

        if isinstance(df, SparkDataFrame):
            return True

        return False

else:
    # Mocks when pyspark is not installed

    class SparkDataFrame:
        pass

    class SparkConnectDataFrame:
        pass

    class SparkColumn:
        pass

    class SparkSession:
        pass

    def is_spark_dataframe(df):
        return True



import os
from pathlib import Path
from typing import Any
from typing import Union

from pydantic import Field
from pydantic import model_validator

from laktory._settings import settings
from laktory.models.basemodel import BaseModel
from laktory.models.resources.baseresource import ResourceLookup
from laktory.models.resources.databricks.accesscontrol import AccessControl
from laktory.models.resources.databricks.permissions import Permissions
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class DbfsFileLookup(ResourceLookup):
    """
    Attributes
    ----------
    path:
        Path on DBFS for the file from which to get content.
    limit_file_size:
        Do not load content for files larger than 4MB.
    """

    path: str = Field(serialization_alias="id")
    limit_file_size: bool = True


class DbfsFile(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks DBFS File

    Attributes
    ----------
    access_controls:
        List of file access controls
    dirpath:
        Workspace directory inside rootpath in which the DBFS file is
        deployed. Used only if `path` is not specified.
    lookup_existing:
        Specifications for looking up existing resource. Other attributes will
        be ignored.
    path:
         DBFS filepath for the file. Overwrite `rootpath` and `dirpath`.
    rootpath:
        Root directory to which all DBFS files are deployed to. Used only if
        `path` is not specified.
    source:
        Path to file on local filesystem.

    Examples
    --------
    ```py
    from laktory import models

    file = models.resources.databricks.DbfsFile(
        source="./data/stock_prices/prices.json",
    )
    print(file.path)
    # > /prices.json

    file = models.resources.databricks.DbfsFile(
        source="./data/stock_prices/prices.json",
        rootpath="/data/",
    )
    print(file.path)
    # > /data/prices.json

    file = models.resources.databricks.DbfsFile(
        source="./data/stock_prices/prices.json",
        rootpath="/data/",
        dirpath="stock_prices/",
    )
    print(file.path)
    # > /data/stock_prices/prices.json
    ```
    """

    access_controls: list[AccessControl] = []
    dirpath: str = None
    lookup_existing: DbfsFileLookup = Field(None, exclude=True)
    path: str = None
    rootpath: str = "/"
    source: str

    @classmethod
    def lookup_defaults(cls) -> dict:
        return {"path": ""}

    @property
    def filename(self) -> str:
        """File filename"""
        return os.path.basename(self.source)

    @model_validator(mode="after")
    def set_paths(self) -> Any:
        # Path set
        if self.path:
            return self

        # root
        if self.rootpath is None:
            self.rootpath = settings.workspace_laktory_root

        # dir
        if self.dirpath is None:
            self.dirpath = ""
        if self.dirpath.startswith("/"):
            self.dirpath = self.dirpath[1:]

        # path
        _path = Path(self.rootpath) / self.dirpath / self.filename
        self.path = _path.as_posix()

        return self

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        """path with special characters `/`, `.`, `\\` replaced with `-`"""
        # key = os.path.splitext(self.path)[0]
        key = self.path
        key = key.replace("/", "-")
        key = key.replace("\\", "-")
        key = key.replace(".", "-")
        for i in range(5):
            if key.startswith("-"):
                key = key[1:]
        return key

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        resources = []
        if self.access_controls:
            resources += [
                Permissions(
                    resource_name=f"permissions-{self.resource_name}",
                    access_controls=self.access_controls,
                    workspace_file_path=f"${{resources.{self.resource_name}.path}}",
                )
            ]
        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:DbfsFile"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["access_controls", "dirpath", "rootpath"]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_dbfs_file"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



from laktory.models.basemodel import BaseModel
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class GroupMember(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks secret ACL

    Attributes
    ----------
    group_id:
        This is the id of the group resource.
    member_id:
        This is the id of the group, service principal, or user.
    """

    group_id: str = None
    member_id: str = None

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        return f"{self.group_id}-{self.member_id}"

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:GroupMember"

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_group_member"



import polars as pl


def union(df: pl.DataFrame, other: pl.DataFrame) -> pl.DataFrame:
    """
    Return a new DataFrame containing the union of rows in this and another
    DataFrame.

    Parameters
    ----------
    df:
        Input DataFrame
    other:
        Other DataFrame

    Examples
    --------
    ```py
    import polars as pl

    import laktory  # noqa: F401

    df0 = pl.DataFrame(
        {
            "symbol": ["AAPL", "AAPL"],
            "price": [200.0, 205.0],
            "tstamp": ["2023-09-01", "2023-09-02"],
        }
    )

    df = df0.laktory.union(df0)
    print(df.glimpse(return_as_string=True))
    '''
    Rows: 4
    Columns: 3
    $ symbol <str> 'AAPL', 'AAPL', 'AAPL', 'AAPL'
    $ price  <f64> 200.0, 205.0, 200.0, 205.0
    $ tstamp <str> '2023-09-01', '2023-09-02', '2023-09-01', '2023-09-02'
    '''
    ```
    """
    return pl.concat([df, other])



from typing import Union

import polars as pl

__all__ = [
    "roundp",
]


# --------------------------------------------------------------------------- #
# Rounding                                                                    #
# --------------------------------------------------------------------------- #


def roundp(
    x: pl.Expr,
    p: Union[float, pl.Expr] = 1.0,
) -> pl.Expr:
    """
    Evenly round to the given precision

    Parameters
    ------
    x:
        Input column
    p:
        Precision

    Returns
    -------
    :
        Output column

    Examples
    --------
    ```py
    import polars as pl

    import laktory  # noqa: F401

    df = pl.DataFrame([[0.781], [13.0]], ["x"])
    df = df.with_columns(y=pl.Expr.laktory.roundp(pl.col("x"), p=5))
    print(df.glimpse(return_as_string=True))
    '''
    Rows: 2
    Columns: 2
    $ x <f64> 0.781, 13.0
    $ y <f64> 0.0, 15.0
    '''

    df = df.with_columns(y=pl.Expr.laktory.roundp(pl.col("x"), p=0.25))
    print(df.glimpse(return_as_string=True))
    '''
    Rows: 2
    Columns: 2
    $ x <f64> 0.781, 13.0
    $ y <f64> 0.75, 13.0
    '''
    ```
    """
    return (x / p).round() * p



from pathlib import Path
from typing import Any
from typing import Union

from pydantic import model_validator

from laktory._settings import settings
from laktory.models.basemodel import BaseModel
from laktory.models.resources.databricks.accesscontrol import AccessControl
from laktory.models.resources.databricks.permissions import Permissions
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class Dashboard(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Lakeview Dashboard

    Attributes
    ----------
    access_controls:
        Dashboard access controls
    dirpath:
        Workspace directory inside rootpath in which the dashboard is deployed.
        Used only if `parent_path` is not specified.
    display_name:
        The display name of the dashboard.
    embed_credentials:
        Whether to embed credentials in the dashboard. Default is true.
    file_path:
        The path to the dashboard JSON file. Conflicts with serialized_dashboard.
    name_prefix:
        Prefix added to the dashboard display name
    name_suffix:
        Suffix added to the dashboard display name
    parent_path:
        The workspace path of the folder containing the dashboard. Includes leading slash and no trailing slash.
        If folder doesn't exist, it will be created.
    rootpath:
        Root directory to which all dashboards are deployed to. Can also be
        configured by settings LAKTORY_WORKSPACE_LAKTORY_ROOT environment
        variable. Default is `/.laktory/`. Used only if `parent_path` is not
        specified.
    serialized_dashboard:
        The contents of the dashboard in serialized string form. Conflicts with file_path.
    warehouse_id:
        The warehouse ID used to run the dashboard.

    Examples
    --------
    ```py
    import io

    from laktory import models

    # Define job
    job_yaml = '''
    display_name: databricks-costs
    file_path: ./dashboards/databricks_costs.json
    parent_path: /.laktory/dashboards
    warehouse_id: a7d9f2kl8mp3q6rt
    access_controls:
        - group_name: account users
          permission_level: CAN_READ
        - group_name: account users
          permission_level: CAN_RUN
    '''
    dashboard = models.resources.databricks.Dashboard.model_validate_yaml(
        io.StringIO(job_yaml)
    )
    ```
    """

    access_controls: list[AccessControl] = []
    dirpath: str = None
    display_name: str
    embed_credentials: bool = None
    file_path: str = None
    name_prefix: str = None
    name_suffix: str = None
    parent_path: str = None
    path: str = None
    rootpath: str = None
    serialized_dashboard: str = None
    warehouse_id: str

    @model_validator(mode="after")
    def set_paths(self) -> Any:
        # Parent Path explicitly set
        if self.parent_path:
            return self

        # root
        if self.rootpath is None:
            self.rootpath = settings.workspace_laktory_root

        # dir
        if self.dirpath is None:
            self.dirpath = ""
        if self.dirpath.startswith("/"):
            self.dirpath = self.dirpath[1:]

        # parent_path
        _path = Path(self.rootpath) / self.dirpath
        self.parent_path = _path.as_posix()

        return self

    @model_validator(mode="after")
    def update_name(self) -> Any:
        if self.name_prefix:
            self.display_name = self.name_prefix + self.display_name
            self.name_prefix = ""
        if self.name_suffix:
            self.display_name = self.display_name + self.name_suffix
            self.name_suffix = ""
        return self

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        return self.display_name

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - permissions
        """
        resources = []
        if self.access_controls:
            resources += [
                Permissions(
                    resource_name=f"permissions-{self.resource_name}",
                    access_controls=self.access_controls,
                    dashboard_id=f"${{resources.{self.resource_name}.id}}",
                )
            ]
        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:Dashboard"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return [
            "access_controls",
            "dirpath",
            "rootpath",
            "name_prefix",
            "name_suffix",
        ]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_dashboard"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes
